//-----------------------------------------------------------------------------
// Yet another on-screen ruler.
// Jonathan Westhues, Mar 2005
//-----------------------------------------------------------------------------
#include <windows.h>
#include <process.h>
#include <commctrl.h>
#include <commdlg.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#define FREEZE_SUBKEY "nruler"
#include "freeze.h"

#define M_PI 3.141592653589

#define SQUARE(x) ((x)*(x))

static HINSTANCE   Instance;

static RECT        DeskRect;

static HWND        PointAWindow;
static HWND        PointBWindow;
static HWND        LineWindow;
static HWND        LineAWindow;
static HWND        LineBWindow;
static HWND        LineCWindow;
static HWND        LineDWindow;

static HWND        InfoWindow;
static HWND        InfoLabels[4];
static HWND        InfoTextboxes[4];

static HBRUSH      InnerBrush;
static HPEN        InnerPen;
static HBRUSH      OuterBrush;
static HPEN        OuterPen;

static HFONT       InfoFont;
static HFONT       SymbolFont;

static BOOL        ShowGeometryWhileDragging = TRUE;
static BOOL        ShowGeometryAlways;
static BOOL        ConstrainLength;
static BOOL        ConstrainAngle;
static BOOL        ConstrainX;
static BOOL        ConstrainY;
static BOOL        OnlyMoveTogether;

static BOOL        GridLinesAlwaysOn;

#define GEOMETRY_FORMAT_REL_BETWEEN_CENTRES 0
#define GEOMETRY_FORMAT_REL_BETWEEN_EDGES   1
#define GEOMETRY_FORMAT_ABSOLUTE            2
static int         GeometryFormat;

static BOOL        InfoWindowLocked;

static double      ScaleFactor = 1.0;

static double      ConstrainedLength;
static double      ConstrainedAngle;

#define MNU_CONSTRAIN_LENGTH        1
#define MNU_CONSTRAIN_ANGLE         2
#define MNU_CONSTRAIN_X             3
#define MNU_CONSTRAIN_Y             4
#define MNU_ONLY_MOVE_TOGETHER      5
#define MNU_SHOW_GEOMETRY           6
#define MNU_SHOW_GEOMETRY_ALWAYS    7
#define MNU_GEO_REL_BETWEEN_CENTRES 8
#define MNU_GEO_REL_BETWEEN_EDGES   9
#define MNU_GEO_ABSOLUTE            10
#define MNU_LINES_HORIZONTAL        11
#define MNU_LINES_VERTICAL          12
#define MNU_LINES_BOTH              13
#define MNU_LINES_NONE              14
#define MNU_LINES_ALWAYS_ON         15
#define MNU_REMOVE_SCALE_FACTOR     16
#define MNU_ABOUT                   17
#define MNU_EXIT                    18

#define MNU_COLOUR_SCHEME_FIRST 20

static const struct {
    char        *name;
    COLORREF     inner;
    COLORREF     outer;
} ColourSchemes[] = {
    { "Black and White",        RGB(255, 255, 255), RGB(  0,   0,   0) },
    { "Black and Blue",         RGB(130, 130, 255), RGB(  0,   0,   0) },
    { "Black and Red",          RGB(255, 130, 130), RGB(  0,   0,   0) },
    { "White and Black",        RGB(  0,   0,   0), RGB(255, 255, 255) },
    { "Blue and Black",         RGB(  0,   0,   0), RGB( 80,  80, 255) },
    { "Red and Black",          RGB(  0,   0,   0), RGB(255,  80,  80) },
    { "Blue and Yellow",        RGB(255, 255, 100), RGB(  0,   0, 255) },
    { "Red and Green",          RGB( 50, 255,  50), RGB(180,   0,   0) },
};
static int SelectedColourScheme = 0;
#define NUM_COLOUR_SCHEMES (sizeof(ColourSchemes)/sizeof(ColourSchemes[0]))


#define END_MARK_WIDTH 16
#define FILL_WIDTH 4

#define INFO_WIDTH  104
#define INFO_HEIGHT 115
#define INFO_MARGIN 20

static enum LinesTag {
    LINES_NONE,
    LINES_HORIZONTAL,
    LINES_VERTICAL,
    LINES_BOTH,
} Lines;

static enum {
    MOVING_NONE,
    MOVING_POINT_A,
    MOVING_POINT_B,
    MOVING_BOTH,
    MOVING_INFO_WINDOW,
} Moving;

static BOOL PlaceInfoNearPointA;

static POINT       PointA, PointB;

// Hook the whenever user attempts to drag something.
static HHOOK       MouseHookHandle;
static int         MouseY;
static int         MouseX;

static void SetUpRulerRegion(void);
static void ChangedPt(void);


void dbp(char *str, ...)
{
    va_list f;
    char buf[1024];
    va_start(f, str);
    vsprintf(buf, str, f);
    OutputDebugString(buf);
    OutputDebugString("\n");
}

//-----------------------------------------------------------------------------
// Round a floating point number to the nearest integer. 0.5 rounds to 1, and
// -0.5 rounds to -1.
//-----------------------------------------------------------------------------
double rint(double x)
{
    if(x > 0) {
        return floor(x + 0.5);
    } else {
        return ceil(x - 0.5);
    }
}

//-----------------------------------------------------------------------------
// Normalize an angle in radians so that it lies in (-pi, pi]
//-----------------------------------------------------------------------------
double AngleDiff(double theta)
{
    while(theta <= (-M_PI)) theta += 2*M_PI;
    while(theta > M_PI) theta -= 2*M_PI;

    return theta;
}

//-----------------------------------------------------------------------------
// Get (dx, dy, r, theta) from the absolute endpoints, in unscaled (pixel)
// units. This is not quite trivial because of the between centres/between
// bounding box corners thing.
//-----------------------------------------------------------------------------
void GetDeltas(int *pDx, int *pDy, int *pR, int *pTheta)
{
    int dx = PointB.x - PointA.x;
    int dy = PointB.y - PointA.y;
    if(GeometryFormat == GEOMETRY_FORMAT_REL_BETWEEN_EDGES) {
        if(dx < 0) {
            dx -= END_MARK_WIDTH;
        } else {
            dx += END_MARK_WIDTH;
        }
        if(dy < 0) {
            dy -= END_MARK_WIDTH;
        } else {
            dy += END_MARK_WIDTH;
        }
    }
    int r = (int)sqrt((double)(SQUARE(dx) + SQUARE(dy)));
    int theta = (int)rint(180*atan2((double)dy, (double)dx)/M_PI);

    if(pDx)     *pDx    = dx;
    if(pDy)     *pDy    = dy;
    if(pR)      *pR     = r;
    if(pTheta)  *pTheta = theta;
}

//-----------------------------------------------------------------------------
// Change the scale factor to make a particular value (dx, dy, r) equal to
// a given quantity that the user typed in.
//-----------------------------------------------------------------------------
static void SetScaleFactor(HWND hwnd)
{
    int raw;
    if(hwnd == InfoTextboxes[0]) {
        GetDeltas(&raw, NULL, NULL, NULL);
    } else if(hwnd == InfoTextboxes[1]) {
        GetDeltas(NULL, &raw, NULL, NULL);
    } else if(hwnd == InfoTextboxes[2]) {
        GetDeltas(NULL, NULL, &raw, NULL);
    }
    
    char buf[20];
    SendMessage(hwnd, WM_GETTEXT, sizeof(buf), (LPARAM)buf);
    double scaled = atof(buf);

    ScaleFactor = scaled / raw;

    ChangedPt();
}

//-----------------------------------------------------------------------------
// Call this after changing the position of one of the points, always. This
// will keep some stats up to date.
//-----------------------------------------------------------------------------
static void ChangedPt(void)
{
    SetUpRulerRegion();
    InvalidateRect(LineWindow, NULL, FALSE);
    InvalidateRect(PointAWindow, NULL, FALSE);
    InvalidateRect(PointBWindow, NULL, FALSE);
    InvalidateRect(InfoWindow, NULL, FALSE);
    InvalidateRect(LineAWindow, NULL, FALSE);
    InvalidateRect(LineBWindow, NULL, FALSE);
    InvalidateRect(LineCWindow, NULL, FALSE);
    InvalidateRect(LineDWindow, NULL, FALSE);

    // Set up the controls in the info window appropriately
    if(GeometryFormat == GEOMETRY_FORMAT_ABSOLUTE) {
        SetWindowText(InfoLabels[0], "X1:");
        SetWindowText(InfoLabels[1], "Y1:");
        SetWindowText(InfoLabels[2], "X2:");
        SetWindowText(InfoLabels[3], "Y2:");
        SendMessage(InfoLabels[3], WM_SETFONT, (WPARAM)InfoFont, TRUE);
        SendMessage(InfoTextboxes[0], EM_SETREADONLY, TRUE, 0);
        SendMessage(InfoTextboxes[1], EM_SETREADONLY, TRUE, 0);
        SendMessage(InfoTextboxes[2], EM_SETREADONLY, TRUE, 0);
    } else {
        SetWindowText(InfoLabels[0], "dX:");
        SetWindowText(InfoLabels[1], "dY:");
        SetWindowText(InfoLabels[2], "r:");
        SetWindowText(InfoLabels[3], "j:");
        SendMessage(InfoLabels[3], WM_SETFONT, (WPARAM)SymbolFont, TRUE);
        SendMessage(InfoTextboxes[0], EM_SETREADONLY, FALSE, 0);
        SendMessage(InfoTextboxes[1], EM_SETREADONLY, FALSE, 0);
        SendMessage(InfoTextboxes[2], EM_SETREADONLY, FALSE, 0);
    }

    char buf[20];
    if(GeometryFormat == GEOMETRY_FORMAT_ABSOLUTE) {
        sprintf(buf, "%d", PointA.x); SetWindowText(InfoTextboxes[0], buf);
        sprintf(buf, "%d", PointA.y); SetWindowText(InfoTextboxes[1], buf);
        sprintf(buf, "%d", PointB.x); SetWindowText(InfoTextboxes[2], buf);
        sprintf(buf, "%d", PointB.y); SetWindowText(InfoTextboxes[3], buf);
    } else {
        int dx, dy, r, theta;
        GetDeltas(&dx, &dy, &r, &theta);

        char *form;

        if(ScaleFactor == 1.0) {
            form = "%.0f";
        } else if(fabs(ScaleFactor) > 1.0) {
            form = "%.1f";
        } else if(fabs(ScaleFactor) < 0.01) {
            form = "%.3f";
        } else {
            form = "%.2f";
        }

        sprintf(buf, form, dx*ScaleFactor);
        SetWindowText(InfoTextboxes[0], buf);
        sprintf(buf, form, dy*ScaleFactor);
        SetWindowText(InfoTextboxes[1], buf);
        sprintf(buf, form, r*fabs(ScaleFactor));
        SetWindowText(InfoTextboxes[2], buf);

        sprintf(buf, "%d", theta);
        SetWindowText(InfoTextboxes[3], buf);
    }

    if(GetForegroundWindow() != PointAWindow) {
        SetForegroundWindow(LineWindow);
        SetForegroundWindow(PointBWindow);
        SetForegroundWindow(PointAWindow);
    }
}

//-----------------------------------------------------------------------------
// Create the brushes and pens that determine our colour scheme.
//-----------------------------------------------------------------------------
static void ApplyColourScheme(void)
{
    COLORREF inner = ColourSchemes[SelectedColourScheme].inner;
    COLORREF outer = ColourSchemes[SelectedColourScheme].outer;

    InnerBrush = CreateSolidBrush(inner);
    InnerPen = CreatePen(PS_SOLID, FILL_WIDTH-2, inner);
    OuterBrush = CreateSolidBrush(outer);
    OuterPen = CreatePen(PS_SOLID, 1, outer);

    ChangedPt();
}

static LRESULT CALLBACK MouseHook(int code, WPARAM wParam, LPARAM lParam)
{
    switch(code) {
        case HC_ACTION: {
            MSLLHOOKSTRUCT *mhs = (MSLLHOOKSTRUCT *)lParam;

            switch(wParam) {
                case WM_MOUSEMOVE: {
                    int dx = MouseX - mhs->pt.x;
                    int dy = MouseY - mhs->pt.y;

                    BOOL ignoreConstraints = FALSE;
                    if(GetAsyncKeyState(VK_CONTROL) & 0x8000) {
                        ignoreConstraints = TRUE;
                    }

                    if(Moving == MOVING_INFO_WINDOW) {
                        RECT r;
                        GetWindowRect(InfoWindow, &r);
                        MoveWindow(InfoWindow, r.left - dx, r.top - dy,
                            INFO_WIDTH, INFO_HEIGHT, TRUE);

                        MouseX -= dx;
                        MouseY -= dy;
                        break;
                    }

                    if(ConstrainLength && !ignoreConstraints) {
                        POINT ptMove, ptFix;

                        if(Moving == MOVING_POINT_A) {
                            ptMove = PointA;
                            ptFix = PointB;
                        } else if(Moving == MOVING_POINT_B) {
                            ptMove = PointB;
                            ptFix = PointA;
                        } else {
                            goto length_already_constrained;
                        }

                        int ly = (ptMove.y - dy) - ptFix.y;
                        int lx = (ptMove.x - dx) - ptFix.x;

                        double theta = atan2((double)ly, (double)lx);

                        int ptax = (int)rint(ConstrainedLength*cos(theta));
                        int ptay = (int)rint(ConstrainedLength*sin(theta));

                        int newx = ptax + ptFix.x;
                        int newy = ptay + ptFix.y;

                        dx = ptMove.x - newx;
                        dy = ptMove.y - newy;
                    }
length_already_constrained:;

                    if(ConstrainAngle && !ignoreConstraints) {
                        POINT ptMove, ptFix;

                        if(Moving == MOVING_POINT_A) {
                            ptMove = PointA;
                            ptFix = PointB;
                        } else if(Moving == MOVING_POINT_B) {
                            ptMove = PointB;
                            ptFix = PointA;
                        } else {
                            goto angle_already_constrained;
                        }

                        int ly = (ptMove.y - dy) - ptFix.y;
                        int lx = (ptMove.x - dx) - ptFix.x;

                        double r = sqrt((double)(SQUARE(lx) + SQUARE(ly)));

                        double thisAngle = atan2((double)ly, (double)lx);
                        double theta = ConstrainedAngle;

                        if(fabs(AngleDiff(theta - thisAngle + M_PI)) <
                           fabs(AngleDiff(theta - thisAngle)))
                        {
                            theta += M_PI;
                        }

                        int ptax = (int)rint(r*cos(theta));
                        int ptay = (int)rint(r*sin(theta));

                        int newx = ptax + ptFix.x;
                        int newy = ptay + ptFix.y;

                        dx = ptMove.x - newx;
                        dy = ptMove.y - newy;
                    }
angle_already_constrained:;

                    if(GetAsyncKeyState(VK_SHIFT) & 0x8000) {
                        POINT ptMove, ptFix;

                        if(Moving == MOVING_POINT_A) {
                            ptMove = PointA; ptFix = PointB;
                        } else if(Moving == MOVING_POINT_B) {
                            ptMove = PointB; ptFix = PointA;
                        } else {
                            goto shift_already_considered;
                        }
                        
                        int ly = (ptMove.y - dy) - ptFix.y;
                        int lx = (ptMove.x - dx) - ptFix.x;

                        double r = sqrt((double)(SQUARE(lx) + SQUARE(ly)));
                        double theta = atan2((double)ly, (double)lx);

                        theta = floor(theta/(M_PI/4) + 0.5)*(M_PI/4);

                        int ptax = (int)rint(r*cos(theta));
                        int ptay = (int)rint(r*sin(theta));

                        int newx = ptax + ptFix.x;
                        int newy = ptay + ptFix.y;

                        dx = ptMove.x - newx;
                        dy = ptMove.y - newy;
                    }
shift_already_considered:;

                    if(ConstrainX && !ignoreConstraints) dx = 0;
                    if(ConstrainY && !ignoreConstraints) dy = 0;

                    if(Moving == MOVING_POINT_A || Moving == MOVING_BOTH) {
                        if(PointA.x - dx < 0) dx = PointA.x;
                        if(PointA.y - dy < 0) dy = PointA.y;

                        if(PointA.x - dx >= DeskRect.right) {
                            dx = PointA.x - DeskRect.right;
                        }
                        if(PointA.y - dy >= DeskRect.bottom) {
                            dy = PointA.y - DeskRect.bottom;
                        }
                    }
                    if(Moving == MOVING_POINT_B || Moving == MOVING_BOTH) {
                        if(PointB.x - dx < 0) dx = PointB.x;
                        if(PointB.y - dy < 0) dy = PointB.y;

                        if(PointB.x - dx >= DeskRect.right) {
                            dx = PointB.x - DeskRect.right;
                        }
                        if(PointB.y - dy >= DeskRect.bottom) {
                            dy = PointB.y - DeskRect.bottom;
                        }
                    }

                    if(Moving == MOVING_POINT_A) {
                        PointA.x -= dx;
                        PointA.y -= dy;
                    } else if(Moving == MOVING_POINT_B) {
                        PointB.x -= dx;
                        PointB.y -= dy;
                    } else if(Moving == MOVING_BOTH) {
                        PointA.x -= dx;
                        PointA.y -= dy;
                        PointB.x -= dx;
                        PointB.y -= dy;
                    } else {
                        // that's odd
                        UnhookWindowsHookEx(MouseHookHandle);
                        break;
                    }

                    if(ignoreConstraints) {
                        int dy = PointB.y - PointA.y;
                        int dx = PointB.x - PointA.x;
                        ConstrainedAngle = atan2((double)dy, (double)dx);
                        ConstrainedLength =
                            sqrt((double)(SQUARE(dx) + SQUARE(dy)));
                    }

                    ChangedPt();

                    MouseX -= dx;
                    MouseY -= dy;
                    break;
                }
                case WM_LBUTTONUP:
                    if(Moving != MOVING_NONE && Moving != MOVING_INFO_WINDOW) {
                        ShowCursor(TRUE);
                    }
                    Moving = MOVING_NONE;
                    UnhookWindowsHookEx(MouseHookHandle);
                    ChangedPt();
                    break;
            }
            break;
        }
    }
    return CallNextHookEx(MouseHookHandle, code, wParam, lParam);
}

//-----------------------------------------------------------------------------
// Set up the non-rectangular window shapes: circles for the two endpoints,
// and a diagonal polygon (long and skinny) for the line.
//-----------------------------------------------------------------------------
static void SetUpRulerRegion(void)
{
    int dfx, dfy;
    if(abs(PointA.x - PointB.x) > abs(PointA.y - PointB.y)) {
        dfx = 0;
        dfy = (FILL_WIDTH/2)+1;
    } else {
        dfx = (FILL_WIDTH/2)+1;
        dfy = 0;
    }

    POINT points[4] = {
        // actual line
        { PointA.x - dfx, PointA.y - dfy },
        { PointA.x + dfx, PointA.y + dfy },
        { PointB.x + dfx, PointB.y + dfy },
        { PointB.x - dfx, PointB.y - dfy},
    };

    HRGN hrgn = CreatePolygonRgn(points, 4, ALTERNATE);
    SetWindowRgn(LineWindow, hrgn, TRUE);

    hrgn = CreateRoundRectRgn(0, 0, END_MARK_WIDTH, END_MARK_WIDTH,
        END_MARK_WIDTH, END_MARK_WIDTH);
    SetWindowRgn(PointAWindow, hrgn, TRUE);

    hrgn = CreateRoundRectRgn(0, 0, END_MARK_WIDTH, END_MARK_WIDTH,
        END_MARK_WIDTH, END_MARK_WIDTH);
    SetWindowRgn(PointBWindow, hrgn, TRUE);

    MoveWindow(PointAWindow,
        PointA.x - (END_MARK_WIDTH/2), PointA.y - (END_MARK_WIDTH/2),
        END_MARK_WIDTH, END_MARK_WIDTH, TRUE);

    MoveWindow(PointBWindow,
        PointB.x - (END_MARK_WIDTH/2), PointB.y - (END_MARK_WIDTH/2),
        END_MARK_WIDTH, END_MARK_WIDTH, TRUE);

    MoveWindow(LineWindow, 0, 0, DeskRect.right, DeskRect.bottom, TRUE);

    if(Lines == LINES_NONE || (Moving == MOVING_NONE && !GridLinesAlwaysOn)) {
        ShowWindow(LineAWindow, SW_HIDE);
        ShowWindow(LineBWindow, SW_HIDE);
        ShowWindow(LineCWindow, SW_HIDE);
        ShowWindow(LineDWindow, SW_HIDE);
    } else {
        if(Lines == LINES_VERTICAL || Lines == LINES_BOTH) {
            int x1, x2;
            x1 = min(PointA.x, PointB.x);
            x2 = max(PointA.x, PointB.x);
            if(GeometryFormat == GEOMETRY_FORMAT_REL_BETWEEN_EDGES) {
                x1 -= (END_MARK_WIDTH/2);
                x2 += (END_MARK_WIDTH/2) - 1;
            }
            MoveWindow(LineAWindow, x1, 0, 1, DeskRect.bottom, TRUE);
            MoveWindow(LineBWindow, x2, 0, 1, DeskRect.bottom, TRUE);

            ShowWindow(LineAWindow, SW_SHOW);
            ShowWindow(LineBWindow, SW_SHOW);
            if(Lines != LINES_BOTH) {
                ShowWindow(LineCWindow, SW_HIDE);
                ShowWindow(LineDWindow, SW_HIDE);
            }
        }

        if(Lines == LINES_HORIZONTAL || Lines == LINES_BOTH) {
            int y1, y2;
            y1 = min(PointA.y, PointB.y);
            y2 = max(PointA.y, PointB.y);
            if(GeometryFormat == GEOMETRY_FORMAT_REL_BETWEEN_EDGES) {
                y1 -= (END_MARK_WIDTH/2);
                y2 += (END_MARK_WIDTH/2) - 1;
            }
            MoveWindow(LineCWindow, 0, y1, DeskRect.right, 1, TRUE);
            MoveWindow(LineDWindow, 0, y2, DeskRect.right, 1, TRUE);

            ShowWindow(LineCWindow, SW_SHOW);
            ShowWindow(LineDWindow, SW_SHOW);
            if(Lines != LINES_BOTH) {
                ShowWindow(LineAWindow, SW_HIDE);
                ShowWindow(LineBWindow, SW_HIDE);
            }
        }
    }

    if(ShowGeometryAlways ||
        (ShowGeometryWhileDragging && Moving != MOVING_NONE))
    {
        BOOL desperate = FALSE;

        int px, py;
        int ox, oy;
        if(PlaceInfoNearPointA) {
            px = PointA.x; py = PointA.y;
            ox = PointB.x; oy = PointB.y;
        } else {
            px = PointB.x; py = PointB.y;
            ox = PointA.x; oy = PointA.y;
        }

try_again:
        int x, y;
        if(py < oy && py > INFO_HEIGHT + INFO_MARGIN) {
            x = px;
            y = py - INFO_MARGIN - INFO_HEIGHT;
        } else if(px < ox && px > INFO_WIDTH + INFO_MARGIN) {
            y = py;
            x = px - INFO_MARGIN - INFO_WIDTH;
        } else if(py > oy && py < DeskRect.bottom - INFO_HEIGHT - INFO_MARGIN) {
            x = px;
            y = py + INFO_MARGIN;
        } else if(px > ox && px < DeskRect.right - INFO_WIDTH - INFO_MARGIN) {
            y = py;
            x = px + INFO_MARGIN;
        } else if(desperate) {
            if(py < (DeskRect.bottom/2)) {
                y = py + INFO_MARGIN;
            } else {
                y = py - INFO_MARGIN;
            }
            if(px < (DeskRect.right/2)) {
                x = px + INFO_MARGIN;
            } else {
                x = px - INFO_MARGIN;
            }
        } else {
            desperate = TRUE;
            int t;
            t = px; px = ox; ox = t;
            t = py; py = oy; oy = t;
            goto try_again;
        }
        x = max(x, 0);
        y = max(y, 0);
        x = min(x, DeskRect.right - INFO_WIDTH);
        y = min(y, DeskRect.bottom - INFO_HEIGHT);

        if(!InfoWindowLocked) {
            MoveWindow(InfoWindow, x, y, INFO_WIDTH, INFO_HEIGHT, TRUE);
        }
        ShowWindow(InfoWindow, SW_SHOW);
    } else {
        InfoWindowLocked = FALSE;
        ShowWindow(InfoWindow, SW_HIDE);
    }
}

//-----------------------------------------------------------------------------
// Popup menu that allows the user to constrain the ruler in certain ways.
//-----------------------------------------------------------------------------
static void ShowPopupMenu(HWND hwnd, int x, int y)
{
    HMENU m = CreatePopupMenu();

    HMENU schemes = CreatePopupMenu();
    int i;
    for(i = 0; i < NUM_COLOUR_SCHEMES; i++) {
        AppendMenu(schemes, MF_STRING, MNU_COLOUR_SCHEME_FIRST+i,  
            ColourSchemes[i].name);
        if(i == SelectedColourScheme) {
            CheckMenuItem(schemes, MNU_COLOUR_SCHEME_FIRST+i, MF_CHECKED);
        }
    }

    HMENU showGeometry = CreatePopupMenu();
    AppendMenu(showGeometry, MF_STRING, MNU_GEO_REL_BETWEEN_CENTRES,
        "Relative, Between &Centres of Endpoint Circles");
    AppendMenu(showGeometry, MF_STRING, MNU_GEO_REL_BETWEEN_EDGES,
        "Relative, Along Diagonal of &Bounding Box");
    AppendMenu(showGeometry, MF_STRING, MNU_GEO_ABSOLUTE,
        "&Absolute (Screen Pixel), Centres of Endpoint Circles");
    if(GeometryFormat == GEOMETRY_FORMAT_REL_BETWEEN_CENTRES) {
        CheckMenuItem(showGeometry, MNU_GEO_REL_BETWEEN_CENTRES, MF_CHECKED);
    } else if(GeometryFormat == GEOMETRY_FORMAT_REL_BETWEEN_EDGES) {
        CheckMenuItem(showGeometry, MNU_GEO_REL_BETWEEN_EDGES, MF_CHECKED);
    } else {
        CheckMenuItem(showGeometry, MNU_GEO_ABSOLUTE, MF_CHECKED);
    }

    HMENU lines = CreatePopupMenu();
    AppendMenu(lines, MF_STRING, MNU_LINES_HORIZONTAL, "&Horizontal Lines");
    AppendMenu(lines, MF_STRING, MNU_LINES_VERTICAL, "&Vertical Lines");
    AppendMenu(lines, MF_STRING, MNU_LINES_BOTH, "&Both Directions");
    AppendMenu(lines, MF_STRING, MNU_LINES_NONE, "&No Lines");
    AppendMenu(lines, MF_SEPARATOR, 0, "");
    AppendMenu(lines, MF_STRING, MNU_LINES_ALWAYS_ON, "&Always Visible");
    if(Lines == LINES_NONE) {
        CheckMenuItem(lines, MNU_LINES_NONE, MF_CHECKED);
    } else if(Lines == LINES_HORIZONTAL) {
        CheckMenuItem(lines, MNU_LINES_HORIZONTAL, MF_CHECKED);
    } else if(Lines == LINES_BOTH) {
        CheckMenuItem(lines, MNU_LINES_BOTH, MF_CHECKED);
    } else {
        CheckMenuItem(lines, MNU_LINES_VERTICAL, MF_CHECKED);
    }
    if(GridLinesAlwaysOn) {
        CheckMenuItem(lines, MNU_LINES_ALWAYS_ON, MF_CHECKED);
    }

    AppendMenu(m, MF_STRING, MNU_SHOW_GEOMETRY,
        "Show &Geometry While Dragging");
    AppendMenu(m, MF_STRING, MNU_SHOW_GEOMETRY_ALWAYS,
        "Show Geometry &Always");
    AppendMenu(m, MF_SEPARATOR, 0, "");
    AppendMenu(m, MF_STRING, MNU_CONSTRAIN_LENGTH, "Constrain &Length");
    AppendMenu(m, MF_STRING, MNU_CONSTRAIN_ANGLE,  "Constrain &Angle");
    AppendMenu(m, MF_SEPARATOR, 0, "");
    AppendMenu(m, MF_STRING, MNU_CONSTRAIN_X,      "Constrain &X Coordinate");
    AppendMenu(m, MF_STRING, MNU_CONSTRAIN_Y,      "Constrain &Y Coordinate");
    AppendMenu(m, MF_SEPARATOR, 0, "");
    AppendMenu(m, MF_STRING, MNU_ONLY_MOVE_TOGETHER, 
        "Both Points Move &Together");
    AppendMenu(m, MF_SEPARATOR, 0, "");
    AppendMenu(m, MF_STRING | MF_POPUP, (UINT)lines, "&Draw Grid Lines");
    AppendMenu(m, MF_STRING | MF_POPUP, (UINT)schemes, "&Colour Scheme");
    AppendMenu(m, MF_STRING | MF_POPUP, (UINT)showGeometry, "Geometry &Format");
    AppendMenu(m, MF_SEPARATOR, 0, "");
    if(ScaleFactor != 1.0) {
        char buf[100];
        if(fabs(ScaleFactor) < 0.01) {
            sprintf(buf, "Remove &Scale Factor [now %.6f]", ScaleFactor);
        } else {
            sprintf(buf, "Remove &Scale Factor [now %.3f]", ScaleFactor);
        }
        AppendMenu(m, MF_STRING, MNU_REMOVE_SCALE_FACTOR, buf);
        AppendMenu(m, MF_SEPARATOR, 0, "");
    }
    AppendMenu(m, MF_STRING, MNU_ABOUT,            "&About");
    AppendMenu(m, MF_SEPARATOR, 0, "");
    AppendMenu(m, MF_STRING, MNU_EXIT,             "E&xit");


    if(ShowGeometryWhileDragging)
        CheckMenuItem(m, MNU_SHOW_GEOMETRY, MF_CHECKED);
    if(ShowGeometryAlways)
        CheckMenuItem(m, MNU_SHOW_GEOMETRY_ALWAYS, MF_CHECKED);
    if(ConstrainLength)
        CheckMenuItem(m, MNU_CONSTRAIN_LENGTH, MF_CHECKED);
    if(ConstrainAngle)
        CheckMenuItem(m, MNU_CONSTRAIN_ANGLE, MF_CHECKED);
    if(ConstrainX)
        CheckMenuItem(m, MNU_CONSTRAIN_X, MF_CHECKED);
    if(ConstrainY)
        CheckMenuItem(m, MNU_CONSTRAIN_Y, MF_CHECKED);
    if(OnlyMoveTogether)
        CheckMenuItem(m, MNU_ONLY_MOVE_TOGETHER, MF_CHECKED);

    int ret = TrackPopupMenuEx(m,
        TPM_NONOTIFY | TPM_RETURNCMD | TPM_NOANIMATION | TPM_RIGHTBUTTON,
        x, y, hwnd, NULL);

    switch(ret) {
        case MNU_SHOW_GEOMETRY:
            ShowGeometryWhileDragging = !ShowGeometryWhileDragging;
            break;

        case MNU_SHOW_GEOMETRY_ALWAYS:
            ShowGeometryAlways = !ShowGeometryAlways;
            ChangedPt();
            break;

        case MNU_CONSTRAIN_LENGTH:
            if(ConstrainLength) {
                ConstrainLength = FALSE;
            } else {
                int dy = PointB.y - PointA.y;
                int dx = PointB.x - PointA.x;
                ConstrainedLength = sqrt((double)(SQUARE(dx) + SQUARE(dy)));
                ConstrainLength = TRUE;

                ConstrainX = FALSE;
                ConstrainY = FALSE;
                ConstrainAngle = FALSE;
                OnlyMoveTogether = FALSE;
            }
            break;

        case MNU_CONSTRAIN_ANGLE:
            if(ConstrainAngle) {
                ConstrainAngle = FALSE;
            } else {
                int dy = PointB.y - PointA.y;
                int dx = PointB.x - PointA.x;
                ConstrainedAngle = atan2((double)dy, (double)dx);
                ConstrainAngle = TRUE;

                ConstrainX = FALSE;
                ConstrainY = FALSE;
                ConstrainLength = FALSE;
                OnlyMoveTogether = FALSE;
            }
            break;

        case MNU_CONSTRAIN_X:
            ConstrainX = !ConstrainX;
            if(ConstrainX) {
                ConstrainY = FALSE;
                ConstrainAngle = FALSE;
                ConstrainLength = FALSE;
                OnlyMoveTogether = FALSE;
            }
            break;

        case MNU_CONSTRAIN_Y:
            ConstrainY = !ConstrainY;
            if(ConstrainY) {
                ConstrainX = FALSE;
                ConstrainAngle = FALSE;
                ConstrainLength = FALSE;
                OnlyMoveTogether = FALSE;
            }
            break;

        case MNU_ONLY_MOVE_TOGETHER:
            OnlyMoveTogether = !OnlyMoveTogether;
            if(OnlyMoveTogether) {
                ConstrainX = FALSE;
                ConstrainY = FALSE;
                ConstrainAngle = FALSE;
                ConstrainLength = FALSE;
            }
            break;

        case MNU_GEO_REL_BETWEEN_CENTRES:
            GeometryFormat = GEOMETRY_FORMAT_REL_BETWEEN_CENTRES;
            ChangedPt();
            break;

        case MNU_GEO_REL_BETWEEN_EDGES:
            GeometryFormat = GEOMETRY_FORMAT_REL_BETWEEN_EDGES;
            ChangedPt();
            break;

        case MNU_GEO_ABSOLUTE:
            ScaleFactor = 1.0;
            GeometryFormat = GEOMETRY_FORMAT_ABSOLUTE;
            ChangedPt();
            break;

        case MNU_LINES_VERTICAL:
            Lines = LINES_VERTICAL;
            ChangedPt();
            break;

        case MNU_LINES_HORIZONTAL:
            Lines = LINES_HORIZONTAL;
            ChangedPt();
            break;

        case MNU_LINES_BOTH:
            Lines = LINES_BOTH;
            ChangedPt();
            break;

        case MNU_LINES_NONE:
            Lines = LINES_NONE;
            ChangedPt();
            break;

        case MNU_LINES_ALWAYS_ON:
            GridLinesAlwaysOn = !GridLinesAlwaysOn;
            ChangedPt();
            break;

        case MNU_REMOVE_SCALE_FACTOR:
            ScaleFactor = 1.0;
            ChangedPt();
            break;

        case MNU_ABOUT:
            MessageBox(PointAWindow,
                "All lengths default to pixels, and all angles are in "
                "degrees. Type a length into the textbox and hit <Enter> "
                "to change the scale factor so that the length you just "
                "typed is now correct. You can use this to scale "
                "off drawings."
                "\r\n\r\n"
                "Drag the ruler or the endpoints. Shift+drag to force "
                "movement along n*45 degree lines. Ctrl+drag to temporarily "
                "ignore constraints (length, angle, etc.)."
                "\r\n\r\n"
                "Copyright Jonathan Westhues 2005. See http://cq.cx/ for more "
                "information.",
                "About nRuler",
                MB_ICONINFORMATION);
            break;

        case MNU_EXIT:
            PostQuitMessage(0);
            break;

        default: {
            int sel = ret - MNU_COLOUR_SCHEME_FIRST;
            if(sel >= 0 && sel < NUM_COLOUR_SCHEMES) {
                SelectedColourScheme = sel;
                ApplyColourScheme();
            }
            break;
        }
    }
}

//-----------------------------------------------------------------------------
// Draw the ruler; outline it and then fill, so that no matter what background
// it's against it will show up.
//-----------------------------------------------------------------------------
static void DrawRuler(HDC hdc)
{
    int dfx, dfy;
    if(abs(PointA.x - PointB.x) > abs(PointA.y - PointB.y)) {
        dfx = 0;
        dfy = (FILL_WIDTH/2);
    } else {
        dfx = (FILL_WIDTH/2);
        dfy = 0;
    }

    RECT r;
    GetClientRect(LineWindow, &r);
    FillRect(hdc, &r, OuterBrush);

    SelectObject(hdc, InnerPen);
    MoveToEx(hdc, PointA.x, PointA.y, NULL);
    LineTo(hdc, PointB.x, PointB.y);
}

//-----------------------------------------------------------------------------
// Draw either of the endpoints; simple circular thingies.
//-----------------------------------------------------------------------------
static LRESULT EndMarkProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
        case WM_PAINT: {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hwnd, &ps);

            RECT r;
            GetClientRect(hwnd, &r);
            FillRect(hdc, &r, OuterBrush);
            r.top += 1; r.left += 1; r.bottom -= 2; r.right -= 2;
            SelectObject(hdc, InnerBrush);
            SelectObject(hdc, OuterPen);
            Ellipse(hdc, r.left, r.top, r.right, r.bottom);

            EndPaint(hwnd, &ps);
            return 1;
        }
        case WM_RBUTTONDOWN: {
            int x = LOWORD(lParam);
            int y = HIWORD(lParam);
            POINT pt;
            pt.x = x; pt.y = y;
            ClientToScreen(hwnd, &pt);

            ShowPopupMenu(hwnd, pt.x, pt.y);
            break;
        }
        case WM_LBUTTONDBLCLK:
        case WM_LBUTTONDOWN: {
            int x = LOWORD(lParam);
            int y = HIWORD(lParam);

            if(hwnd == PointAWindow) {
                PlaceInfoNearPointA = TRUE;
            } else {
                PlaceInfoNearPointA = FALSE;
            }

            if(msg == WM_LBUTTONDBLCLK || OnlyMoveTogether) {
                Moving = MOVING_BOTH;
            } else if(hwnd == PointAWindow) {
                Moving = MOVING_POINT_A;
            } else if(hwnd == PointBWindow) {
                Moving = MOVING_POINT_B;
            } else {
                Moving = MOVING_NONE;
                break;
            }

            POINT pt;
            pt.x = x; pt.y = y;
            ClientToScreen(hwnd, &pt);
            MouseX = pt.x;
            MouseY = pt.y;
            ShowCursor(FALSE);
            MouseHookHandle = SetWindowsHookEx(WH_MOUSE_LL,
                    (HOOKPROC)MouseHook, Instance, 0);
            break;
        }
        case WM_CLOSE:
        case WM_DESTROY:
            PostQuitMessage(0);
            break;

        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }

    return 1;
}

//-----------------------------------------------------------------------------
// Draw the line connecting the endpoints.
//-----------------------------------------------------------------------------
static LRESULT LineProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
        case WM_RBUTTONDOWN: {
            int x = LOWORD(lParam);
            int y = HIWORD(lParam);
            POINT pt;
            pt.x = x; pt.y = y;
            ClientToScreen(hwnd, &pt);

            SetForegroundWindow(PointBWindow);
            SetForegroundWindow(PointAWindow);
            ShowPopupMenu(hwnd, pt.x, pt.y);
            break;
        }
        case WM_LBUTTONDBLCLK:
        case WM_LBUTTONDOWN: {
            int x = LOWORD(lParam);
            int y = HIWORD(lParam);

            Moving = MOVING_BOTH;

            POINT pt;
            pt.x = x; pt.y = y;
            ClientToScreen(hwnd, &pt);
            MouseX = pt.x;
            MouseY = pt.y;
            ShowCursor(FALSE);
            MouseHookHandle = SetWindowsHookEx(WH_MOUSE_LL,
                    (HOOKPROC)MouseHook, Instance, 0);
            ChangedPt();
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hwnd, &ps);

            DrawRuler(hdc);

            EndPaint(hwnd, &ps);
            return 1;
        }
        case WM_CLOSE:
        case WM_DESTROY:
            PostQuitMessage(0);
            break;

        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }

    return 1;
}

//-----------------------------------------------------------------------------
// Draw horizontal or vertical grid lines; easy, just make the window one
// pixel wide/tall and fill with solid.
//-----------------------------------------------------------------------------
static LRESULT GridLineProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
        case WM_PAINT: {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hwnd, &ps);

            RECT r;
            GetClientRect(hwnd, &r);
            if(r.right > r.bottom) {
                int i;
                for(i = 0; i < r.right; i += 10) {
                    RECT f;
                    f.top = 0; f.bottom = 1;

                    f.left = i; f.right = i + 5;
                    FillRect(hdc, &f, OuterBrush);
                    f.left = i + 5; f.right = i + 10;
                    FillRect(hdc, &f, InnerBrush);
                }
            } else {
                int i;
                for(i = 0; i < r.bottom; i += 10) {
                    RECT f;
                    f.left = 0; f.right = 1;

                    f.top = i; f.bottom = i + 5;
                    FillRect(hdc, &f, OuterBrush);
                    f.top = i + 5; f.bottom = i + 10;
                    FillRect(hdc, &f, InnerBrush);
                }
            }

            EndPaint(hwnd, &ps);
            return 1;
        }

        case WM_CLOSE:
        case WM_DESTROY:
            PostQuitMessage(0);
            break;

        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }

    return 1;
}

//-----------------------------------------------------------------------------
// Draw the `where are we' information on-screen.
//-----------------------------------------------------------------------------
static LRESULT InfoProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
        case WM_RBUTTONDOWN: {
            int x = LOWORD(lParam);
            int y = HIWORD(lParam);
            POINT pt;
            pt.x = x; pt.y = y;
            ClientToScreen(hwnd, &pt);

            ShowPopupMenu(hwnd, pt.x, pt.y);
            break;
        }
        case WM_LBUTTONDOWN: {
            int x = LOWORD(lParam);
            int y = HIWORD(lParam);

            Moving = MOVING_INFO_WINDOW;
            InfoWindowLocked = TRUE;

            POINT pt;
            pt.x = x; pt.y = y;
            ClientToScreen(hwnd, &pt);
            MouseX = pt.x;
            MouseY = pt.y;
            MouseHookHandle = SetWindowsHookEx(WH_MOUSE_LL,
                    (HOOKPROC)MouseHook, Instance, 0);
            break;
        }
        case WM_CLOSE:
        case WM_DESTROY:
            PostQuitMessage(0);
            break;

        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }

    return 1;
}

LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    if(hwnd == LineWindow) {
        return LineProc(hwnd, msg, wParam, lParam);
    } else if(hwnd == PointAWindow || hwnd == PointBWindow) {
        return EndMarkProc(hwnd, msg, wParam, lParam);
    } else if(hwnd == LineAWindow || hwnd == LineBWindow 
           || hwnd == LineCWindow || hwnd == LineDWindow)
    {
        return GridLineProc(hwnd, msg, wParam, lParam);
    } else if(hwnd == InfoWindow) {
        return InfoProc(hwnd, msg, wParam, lParam);
    } else {
        return DefWindowProc(hwnd, msg, wParam, lParam);
    }
}

//-----------------------------------------------------------------------------
// Create our window class; nothing exciting.
//-----------------------------------------------------------------------------
static BOOL MakeWindowClass()
{
    WNDCLASSEX wc;
    memset(&wc, 0, sizeof(wc));
    wc.cbSize = sizeof(wc);

    wc.style            = CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW | CS_OWNDC |
                            CS_DBLCLKS;
    wc.lpfnWndProc      = (WNDPROC)MainWndProc;
    wc.hInstance        = Instance;
    wc.hbrBackground    = (HBRUSH)COLOR_BTNSHADOW;
    wc.lpszClassName    = "nruler";
    wc.lpszMenuName     = NULL;
    wc.hCursor          = LoadCursor(NULL, IDC_SIZEALL);

    return RegisterClassEx(&wc);
}

//-----------------------------------------------------------------------------
// Entry point into the program.
//-----------------------------------------------------------------------------
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
    LPSTR lpCmdLine, INT nCmdShow)
{
    Instance = hInstance;

    InfoFont = CreateFont(16, 0, 0, 0, FW_REGULAR, FALSE, FALSE, FALSE,
        ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
        ANTIALIASED_QUALITY, FF_DONTCARE, "Tahoma");

    SymbolFont = CreateFont(16, 0, 0, 0, FW_REGULAR, FALSE, FALSE, FALSE,
        ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
        ANTIALIASED_QUALITY, FF_DONTCARE, "Symbol");

    PointA.x = 300;
    PointA.y = 300;
    PointB.x = 500;
    PointB.y = 600;

    MakeWindowClass();

    GetClientRect(GetDesktopWindow(), &DeskRect);

    InfoWindow = CreateWindowEx(WS_EX_TOPMOST | WS_EX_TOOLWINDOW, "nruler", "",
        WS_POPUP | WS_DLGFRAME, 0, 0, INFO_WIDTH, INFO_HEIGHT, NULL, NULL,
        Instance, NULL);

    int i;
    for(i = 0; i < 4; i++) {
        InfoLabels[i] = CreateWindowEx(0, WC_STATIC, "Top:",
            WS_CHILD | WS_CLIPSIBLINGS | WS_VISIBLE | SS_RIGHT,
            5, 5 + 26*i, 22, 21, InfoWindow, NULL, Instance, NULL);

        InfoTextboxes[i] = CreateWindowEx(WS_EX_CLIENTEDGE, WC_EDIT, "",
            WS_CHILD | ES_AUTOHSCROLL | WS_TABSTOP | ES_READONLY |
                WS_CLIPSIBLINGS | WS_VISIBLE,
            32, 5 + 26*i, 60, 21, InfoWindow, NULL, Instance, NULL);

        SendMessage(InfoLabels[i], WM_SETFONT, (WPARAM)InfoFont, TRUE);
        SendMessage(InfoTextboxes[i], WM_SETFONT, (WPARAM)InfoFont, TRUE);
    }

#define MAKE_WINDOW() \
    CreateWindowEx(WS_EX_TOPMOST | WS_EX_TOOLWINDOW, "nruler", \
        "", WS_POPUP, 0, 0, \
        INFO_WIDTH, INFO_HEIGHT, NULL, NULL, Instance, NULL)
    LineWindow      = MAKE_WINDOW();
    PointAWindow    = MAKE_WINDOW();
    PointBWindow    = MAKE_WINDOW();
    LineAWindow     = MAKE_WINDOW();
    LineBWindow     = MAKE_WINDOW();
    LineCWindow     = MAKE_WINDOW();
    LineDWindow     = MAKE_WINDOW();

    ThawDWORD(PointA.x);
    ThawDWORD(PointA.y);
    ThawDWORD(PointB.x);
    ThawDWORD(PointB.y);
    ThawDWORD(ShowGeometryWhileDragging);
    ThawDWORD(ShowGeometryAlways);
    ThawDWORD(SelectedColourScheme);
    ThawDWORD(GeometryFormat);
    ThawDWORD(GridLinesAlwaysOn);
    DWORD GridLines = (DWORD)Lines; ThawDWORD(GridLines);
    Lines = (enum LinesTag)GridLines;

    ApplyColourScheme();

    ChangedPt();

    ShowWindow(LineWindow, SW_SHOW);
    ShowWindow(PointAWindow, SW_SHOW);
    ShowWindow(PointBWindow, SW_SHOW);

    MSG msg;
    DWORD ret;
    while(ret = GetMessage(&msg, NULL, 0, 0)) {
        if(msg.message == WM_KEYDOWN &&
            (msg.hwnd == InfoTextboxes[0] ||
             msg.hwnd == InfoTextboxes[1] ||
             msg.hwnd == InfoTextboxes[2]))
        {
            if(msg.wParam == VK_RETURN) {
                SetScaleFactor(msg.hwnd);
                continue;
            }
        }

        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    FreezeDWORD(PointA.x);
    FreezeDWORD(PointA.y);
    FreezeDWORD(PointB.x);
    FreezeDWORD(PointB.y);
    FreezeDWORD(ShowGeometryWhileDragging);
    FreezeDWORD(ShowGeometryAlways);
    FreezeDWORD(SelectedColourScheme);
    FreezeDWORD(GeometryFormat);
    FreezeDWORD(GridLinesAlwaysOn);
    GridLines = (DWORD)Lines; FreezeDWORD(GridLines);
    return 0;
}
